from AsyncPQ import ASPQ_Wrapper, asQueue
from ASPQ_Wrapper import AsyncRWQueue
from asQueue import AsyncQueue

class AsyncReadWriteQueue(AsyncQueue):

	def __init__(self, classDict = {}):
		
		super().__init__(classDict)
		self.asyncQueue = AsyncRWQueue()
		self.WRITE_PRIORITY = 1
		self.READ_PRIORITY = 0

	def read(self, operation, *opargs):
		self.asyncQueue.addQUnit(operation, opargs, self.READ_PRIORITY)

	def write(self, operation, *opargs):
		self.asyncQueue.addQUnit(operation, opargs, self.WRITE_PRIORITY)

	def cdRead(self, className, operation, *opargs): 

		classInst = self.classDict[className]
		functionRef = getattr(classInst, operation)
		self.asyncQueue.addQUnit(functionRef, opargs, self.READ_PRIORITY)

	def cdWrite(self, className, operation, *opargs): 

		classInst = self.classDict[className]
		functionRef = getattr(classInst, operation)
		self.asyncQueue.addQUnit(functionRef, opargs, self.WRITE_PRIORITY)



from AsyncPQ import ASPQ_Wrapper, asQueue
from ASPQ_Wrapper import AsyncPQueue
from asQueue import AsyncQueue

class AsyncPriorityQueue(AsyncQueue):

	def __init__(self, classDict = {}):
		
		super().__init__(classDict)
		self.asyncQueue = AsyncPQueue()

	def addTask(self, priority, functionRef, *funcArgs):
		self.asyncQueue.addQUnit(functionRef, funcArgs, priority)

	def addClassTask(self, priority, className, operation, *opArgs): 

		classInst = self.classDict[className]
		functionRef = getattr(classInst, operation)
		self.asyncQueue.addQUnit(functionRef, opArgs, priority)




from setuptools import setup

setup(
  name = 'AsyncPQ',
  packages = ['AsyncPQ'], # this must be the same as the name above
  version = '0.41',
  description = 'Low overhead, dead simple Asyncronous Priority Queue',
  author = 'Michael Usachenko',
  author_email = 'meu2@illinois.edu',
  url = 'https://github.com/themichaelusa/AsyncPQ', # use the URL to the github repo
  download_url = 'https://github.com/themichaelusa/AsyncPQ/archive/0.41.tar.gz', # I'll explain this in a second
  keywords = ['async', 'priority', 'queue'], # arbitrary keywords
  classifiers = [],
)



class AsyncQueue(object):

	def __init__(self, classDict = {}):
		
		self.asyncQueue = None
		self.classDict = classDict

	def updateClassDict(self, newDict):
		self.classDict.update(newDict)

	def processTasks(self):
		return self.asyncQueue.processUnits()


import asyncio

class AsyncPQWrapper(object):

	def __init__(self):
		self.queue = []

	class Unit(object):
	
		def __init__(self, functionRef, fArgs, priority):
			
			self.funcRef = functionRef
			self.funcArgs = fArgs
			self.priority = priority

	def addQUnit(self, functionRef, fArgs, priority):
		self.queue.append(self.Unit(functionRef, fArgs, priority))

	def preprocessUnits(self):

		units = [self.queue.pop() for u in range(len(self.queue))]
		if (len(units) == 0): return None
		taskloop = asyncio.new_event_loop()
		asyncio.set_event_loop(taskloop)
		return (units, taskloop)

	def processUnits(self): pass

class AsyncRWQueue(AsyncPQWrapper):
	
	def __init__(self): 
		super().__init__()
		self.WRITE_PRIORITY = 1
		self.READ_PRIORITY = 0

	def processUnits(self):

		units, taskloop = self.preprocessUnits()
		wUnits = list(filter(lambda x: x.priority == self.WRITE_PRIORITY, units))
		rUnits = list(filter(lambda x: x.priority == self.READ_PRIORITY, units))
		wTasks, rTasks = [], []

		for unit in wUnits:
			wTasks.append(asyncio.ensure_future(unit.funcRef(*unit.funcArgs)))
		for unit in rUnits:
			rTasks.append(asyncio.ensure_future(unit.funcRef(*unit.funcArgs)))

		wFuture, rFuture = asyncio.gather(*wTasks), asyncio.gather(*rTasks)
		taskloop.run_until_complete(wFuture)
		rValues = taskloop.run_until_complete(rFuture)
		taskloop.stop()
		taskloop.close()

		if (len(rValues) == 1): return rValues[0]
		else: return tuple(rValues[::-1])
		
class AsyncPQueue(AsyncPQWrapper):
	
	def __init__(self): super().__init__()

	def processUnits(self):

		units, taskloop = self.preprocessUnits()
		uPriorList = sorted(list(set([unit.priority for unit in units])))[::-1]
		sortedUnits = [list(filter(lambda x: x.priority == u, units)) for u in uPriorList]
		
		sFutures, rValues = [], []
		for sUnit in sortedUnits:
			sortedTasks = []
			for unit in sUnit:
				sortedTasks.append(asyncio.ensure_future(unit.funcRef(*unit.funcArgs)))
			sFutures.append(asyncio.gather(*sortedTasks))
		
		for sFuture in sFutures:
			rValues.append(taskloop.run_until_complete(sFuture))

		taskloop.stop()
		taskloop.close()	
		